@@ -3,8 +3,8 @@ import threading
 import logging
 from functools import partial
 
-from concurrent import futures
 from six.moves.queue import Queue, Empty
+import six
 from threading import Thread
 
 import koji
@@ -12,13 +12,26 @@ from more_executors import Executors
 from more_executors.futures import f_map
 
 from ..source import Source
-from ..model import RpmPushItem, ModuleMdPushItem
-from ..helpers import list_argument, try_int
+from ..model import (
+    KojiBuildInfo,
+    RpmPushItem,
+    ModuleMdPushItem,
+    ModuleMdSourcePushItem,
+    OperatorManifestPushItem,
+    ContainerImagePushItem,
+    SourceContainerImagePushItem,
+)
+from ..helpers import list_argument, try_int, as_completed_with_timeout_reset
 from .modulemd import Module
+from .koji_containers import ContainerArchiveHelper, MIME_TYPE_MANIFEST_LIST
 
 LOG = logging.getLogger("pushsource")
 CACHE_LOCK = threading.RLock()
 
+# Arguments used for retry policy.
+# Provided so it can be overridden from tests to reduce time spent on retries.
+RETRY_ARGS = {}
+
 
 class ListArchivesCommand(object):
     def __init__(self, build):
@@ -104,6 +117,7 @@ class KojiSource(Source):
         rpm=None,
         module_build=None,
         module_filter_filename=None,
+        container_build=None,
         signing_key=None,
         basedir=None,
         threads=4,
@@ -143,6 +157,12 @@ class KojiSource(Source):
 
                 Has no effect if ``module_build`` is not provided.
 
+            container_build (list[str, int])
+                Build identifier(s). Can be koji IDs (integers) or build NVRs.
+                The source will yield all container images produced by these
+                build(s), along with associated artifacts such as operator manifest
+                bundles.
+
             signing_key (list[str])
                 GPG signing key ID(s). If provided, content must be signed
                 using one of the provided keys. Include ``None`` if unsigned
@@ -176,13 +196,10 @@ class KojiSource(Source):
         self._url = url
         self._rpm = [try_int(x) for x in list_argument(rpm)]
         self._module_build = [try_int(x) for x in list_argument(module_build)]
-        self._module_filter_filename = (
-            # retain None values so we can differentiate between
-            # "caller asked for empty filter" vs "caller did not ask for any filter"
-            list_argument(module_filter_filename)
-            if module_filter_filename is not None
-            else None
+        self._module_filter_filename = list_argument(
+            module_filter_filename, retain_none=True
         )
+        self._container_build = [try_int(x) for x in list_argument(container_build)]
         self._signing_key = list_argument(signing_key)
         self._dest = list_argument(dest)
         self._timeout = timeout
@@ -190,9 +207,23 @@ class KojiSource(Source):
         self._cache = {} if cache is None else cache
         self._threads = threads
         self._executor = (
-            executor or Executors.thread_pool(max_workers=threads).with_retry()
+            executor
+            or Executors.thread_pool(name="pushsource-koji", max_workers=threads)
+            .with_retry(**RETRY_ARGS)
+            .with_cancel_on_shutdown()
         )
 
+        self._on_shutdown = []
+        if not executor:
+            self._on_shutdown.append(lambda: self._executor.shutdown(True))
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        for cb in self._on_shutdown:
+            cb()
+
     @property
     def _koji_session(self):
         # A koji client session.
@@ -204,6 +235,28 @@ class KojiSource(Source):
                 tls.koji_session = koji.ClientSession(self._url)
         return tls.koji_session
 
+    def _koji_check(self):
+        # Do a basic connection check with koji.
+        # If this succeeds, we can be reasonably sure that the koji connection is
+        # more or less working.
+        try:
+            version = self._executor.submit(self._koji_get_version).result()
+        except Exception as ex:  # pylint: disable=broad-except
+            # TODO: drop this log when py2 support is dropped.
+            # It's only here because py2 has no exception chaining.
+            msg = "Communication error with koji at %s" % self._url
+            LOG.exception(msg)
+
+            six.raise_from(RuntimeError(msg), ex)
+
+        LOG.debug("Connected to koji %s at %s", version, self._url)
+
+    def _koji_get_version(self):
+        with CACHE_LOCK:
+            if "koji_version" not in self._cache:
+                self._cache["koji_version"] = self._koji_session.getKojiVersion()
+            return self._cache["koji_version"]
+
     def _get_rpm(self, rpm):
         return self._cache["rpm"][rpm]
 
@@ -319,7 +372,7 @@ class KojiSource(Source):
 
         if not meta:
             message = "Module build not found in koji: %s" % nvr
-            LOG.error(message)
+            LOG.debug(message)
             raise ValueError(message)
 
         build_id = meta["id"]
@@ -337,13 +390,137 @@ class KojiSource(Source):
 
             name = self._get_module_name(nvr, file_path)
 
+            klass = ModuleMdPushItem
+            if os.path.basename(file_path) == "modulemd.src.txt":
+                klass = ModuleMdSourcePushItem
+
             out.append(
-                ModuleMdPushItem(
-                    name=name, src=file_path, dest=self._dest, build=meta["nvr"]
-                )
+                klass(name=name, src=file_path, dest=self._dest, build=meta["nvr"])
             )
         return out
 
+    def _push_items_from_container_build(self, nvr, meta):
+        LOG.debug("Looking for container images on %s, %s", nvr, meta)
+
+        if not meta:
+            message = "Container image build not found in koji: %s" % nvr
+            LOG.debug(message)
+            raise ValueError(message)
+
+        # The metadata we are interested in is documented here:
+        # https://github.com/containerbuildsystem/atomic-reactor/blob/061d92e63cf27ae030e8ceed388ec34f51afb17b/docs/koji.md#type-specific-metadata
+        extra = meta.get("extra") or {}
+        typeinfo = extra.get("typeinfo") or {}
+        # Per above doc, it is preferred to use the metadata under 'typeinfo' if present
+        image = typeinfo.get("image") or extra.get("image")
+
+        if image is None:
+            message = "Build %s not recognized as a container image build" % nvr
+            LOG.debug(message)
+            raise ValueError(message)
+
+        build_id = meta["id"]
+
+        archives = self._get_archives(build_id)
+
+        image_archives = [
+            elem
+            for elem in archives
+            if elem.get("btype") == "image" and elem.get("type_name") == "tar"
+        ]
+
+        media_types = image.get("media_types") or []
+        if MIME_TYPE_MANIFEST_LIST in media_types:
+            # in manifest list case it is OK to have multiple archives (different arches).
+            pass
+        elif len(image_archives) != 1:
+            # If we don't have a manifest list, we expect to have an image for one arch.
+            # If there's not exactly one, it's unclear what's going on and we'd better halt.
+            message = (
+                "Could not find (exactly) one container image archive on koji build %s"
+                % nvr
+            )
+            LOG.debug(message)
+            raise ValueError(message)
+
+        out = []
+
+        for archive in image_archives:
+            path = self._pathinfo.typedir(meta, archive["btype"])
+            item_src = os.path.join(path, archive["filename"])
+
+            helper = ContainerArchiveHelper(meta, archive)
+
+            klass = ContainerImagePushItem
+            if image.get("sources_for_nvr"):
+                # This is a source image.
+                klass = SourceContainerImagePushItem
+
+            out.append(
+                klass(
+                    # TODO: name might be changed once we start parsing
+                    # the metadata from atomic-reactor.
+                    name=archive["filename"],
+                    dest=self._dest,
+                    src=item_src,
+                    build=nvr,
+                    # Note, we should be able to use the default KojiBuild
+                    # construction from NVR here. The reason we don't is that
+                    # there is some "impossible" test data in Pub which has
+                    # brew builds whose NVR does not match their (n, v, r)
+                    # in the build data and it's quite tough to fix that now.
+                    build_info=KojiBuildInfo(
+                        name=meta["name"],
+                        version=meta["version"],
+                        release=meta["release"],
+                    ),
+                    source_tags=helper.source_tags,
+                    labels=helper.labels,
+                    arch=helper.arch,
+                    pull_info=helper.pull_info,
+                )
+            )
+
+        # If this build had any operator-manifests archive, add that too.
+        operator = self._get_operator_item(nvr, meta, archives)
+        if operator:
+            out.append(operator)
+
+        return out
+
+    def _get_operator_item(self, nvr, meta, archives):
+        extra = meta.get("extra") or {}
+        typeinfo = extra.get("typeinfo") or {}
+        operator_manifests = typeinfo.get("operator-manifests") or {}
+        archive_name = operator_manifests.get("archive")
+
+        if not archive_name:
+            # Try legacy form
+            archive_name = extra.get("operator_manifests_archive")
+
+        if not archive_name:
+            # No operator manifests on this build
+            return
+
+        operator_archive = [a for a in archives if a["filename"] == archive_name]
+        if len(operator_archive) != 1:
+            message = (
+                "koji build %s metadata refers to missing operator-manifests "
+                'archive "%s"'
+            ) % (nvr, archive_name)
+            raise ValueError(message)
+
+        archive = operator_archive[0]
+        path = self._pathinfo.typedir(meta, archive["btype"])
+        item_src = os.path.join(path, archive["filename"])
+
+        return OperatorManifestPushItem(
+            name=os.path.join(nvr, archive_name),
+            dest=self._dest,
+            src=item_src,
+            build=nvr,
+        )
+
     def _rpm_futures(self):
         # Get info from each requested RPM.
         rpm_fs = [(self._executor.submit(self._get_rpm, rpm), rpm) for rpm in self._rpm]
@@ -366,6 +543,19 @@ class KojiSource(Source):
             for f, build in module_build_fs
         ]
 
+    def _container_futures(self):
+        # Get info from each requested container build.
+        build_fs = [
+            (self._executor.submit(self._get_build, build), build)
+            for build in self._container_build
+        ]
+
+        # Convert them to lists of push items
+        return [
+            f_map(f, partial(self._push_items_from_container_build, build))
+            for f, build in build_fs
+        ]
+
     def _do_fetch(self, koji_queue, exceptions):
         try:
             done = False
@@ -409,6 +599,22 @@ class KojiSource(Source):
             raise
 
     def __iter__(self):
+        # Try a (blocking) call to koji before anything else.
+        #
+        # The reasons for this are:
+        #
+        # - give an error which is hopefully easier to understand if the
+        #   koji connection is completely broken (e.g. caller provides bogus
+        #   URL)
+        #
+        # - work around pyasn1 bug https://github.com/etingof/pyasn1/issues/53
+        #   which affects some (ancient) environments. That thread-safety bug
+        #   can cause parsing of subjectAltName to fail if it first occurs from
+        #   multiple threads simultaneously. Doing a warm-up from one thread
+        #   before we spawn multiple threads is a way to avoid this.
+        #
+        self._koji_check()
+
         # Queue holding all requests we need to make to koji.
         # We try to fetch as much as we can early to make efficient use
         # of multicall.
@@ -423,6 +629,10 @@ class KojiSource(Source):
         for build_id in self._module_build:
             koji_queue.put(GetBuildCommand(ident=build_id, list_archives=True))
 
+        # We'll need to obtain all container image builds
+        for build_id in self._container_build:
+            koji_queue.put(GetBuildCommand(ident=build_id, list_archives=True))
+
         # Put some threads to work on the queue.
         fetch_exceptions = []
         fetch_threads = [
@@ -437,6 +647,23 @@ class KojiSource(Source):
         # Wait for all fetches to finish
         for t in fetch_threads:
             t.start()
+
+        # If we are asked to shut down early then also shut down those threads
+        # as soon as we can.
+        def fetch_shutdown():
+            try:
+                # Empty the queue
+                while True:
+                    koji_queue.get_nowait()
+            except Empty:
+                pass
+
+            # Then ask the threads to join
+            for t in fetch_threads:
+                t.join()
+
+        self._on_shutdown.append(fetch_shutdown)
+
         for t in fetch_threads:
             t.join(self._timeout)
 
@@ -448,9 +675,13 @@ class KojiSource(Source):
         # The queue must be empty now
         assert koji_queue.empty()
 
-        push_items_fs = self._modulemd_futures() + self._rpm_futures()
+        push_items_fs = (
+            self._modulemd_futures() + self._rpm_futures() + self._container_futures()
+        )
 
-        completed_fs = futures.as_completed(push_items_fs, timeout=self._timeout)
+        completed_fs = as_completed_with_timeout_reset(
+            push_items_fs, timeout=self._timeout
+        )
         for f in completed_fs:
             # If an exception occurred, this is where it will be raised.
             for pushitem in f.result():
