@@ -25,17 +25,62 @@ class SourceUrlError(ValueError):
     """
 
 
+class SourceWrapper(object):
+    # Internal class to ensure that all source instances support enter/exit
+    # for with statements even if underlying instance doesn't implement it
+    # (since it originally was not mandatory).
+    def __init__(self, delegate):
+        self.__delegate = delegate
+
+    def __iter__(self):
+        return self.__delegate.__iter__()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        pass
+
+    @classmethod
+    def _maybe_wrap(cls, delegate):
+        if hasattr(delegate, "__enter__"):
+            # no wrapping needed
+            return delegate
+        # wrap with a no-op enter/exit
+        return cls(delegate)
+
+
 class Source(object):
     """A source of push items.
 
     This base class defines the interface for all pushsource backends.
     Instances of a specific backend can be obtained using
     the :meth:`~pushsource.Source.get` method.
+
+    Though not mandatory, instances of ``Source`` are preferably used
+    via ``with`` statements to ensure that all resources are cleaned up
+    when no longer needed, as in example:
+
+    .. code-block:: python
+
+        with Source.get('some-url') as source:
+            for item in source:
+                do_something(item)
+
+    Note that the items produced by a source are not bound to the lifecycle
+    of a source instance, so it's safe to store them and continue using
+    them beyond the end of the ``with`` statement.
     """
 
     _BACKENDS = {}
     _BACKENDS_BUILTIN = {}
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        pass
+
     def __iter__(self):
         """Iterate over the push items contained within this source.
 
@@ -134,7 +179,13 @@ class Source(object):
 
         sig = getfullargspec(klass)
 
-        if "url" in sig.args and parsed.path is not query:
+        # We need to know whether the source backend accepts a 'url' argument,
+        # in which case the next block should kick in. If the backend is a partial
+        # created by us, this info may be available in __pushsource_accepts_url.
+        # See commentary a bit later where this is set.
+        accepts_url = getattr(klass, "__pushsource_accepts_url", "url" in sig.args)
+
+        if accepts_url and parsed.path is not query:
             # If the source accepts a url argument, then the 'path' part
             # of the URL we were provided is itself required to be a URL.
             #
@@ -159,7 +210,19 @@ class Source(object):
         def partial_source(*inner_args, **inner_kwargs):
             kwargs = url_kwargs.copy()
             kwargs.update(inner_kwargs)
-            return klass(*inner_args, **kwargs)
+            return SourceWrapper._maybe_wrap(klass(*inner_args, **kwargs))
+
+        # If the source accepts a 'url' argument, that affects how source
+        # URLs are parsed, as described in the "Implementing a backend"
+        # docs. This must be carried through when partially binding a source.
+        # Stash it in this attribute which we can read back later.
+        #
+        # TODO: please drop this when python2 goes away.
+        # In modern versions of python, functools.wraps already propagates
+        # the signature of the wrapped function, so it is likely possible
+        # to drop this and slightly change above code to use inspect.signature,
+        # then it will "just work".
+        setattr(partial_source, "__pushsource_accepts_url", accepts_url)
 
         return partial_source
 
